/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "pch.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include "Client.h"
#include "Server.h"
#include "GlobalState.h"
#include "HttpRequestMetadata.h"
#include "HttpResponseMetadata.h"
#include "HttpRequestLine.h"
#include "RequestPtr.h"
#include "RequestTag.h"
#include "ErrorResponse.h"
#include "ErrorDescriptor.h"
#include "ErrorCodes.h"
#include "SplittableBuffer.h"
#include "WeakPtr.h"
#include "DataChunk.h"
#include "ScopedIncDec.h"
#include "AbstractDataConsumer.h"
#include "AbstractRequestHandler.h"
#include "AbstractResponseHandler.h"
#include "ServiceContext.h"
#include "HttpStreamWriter.h"
#include "HttpMessageShaper.h"
#include "HttpVersion.h"
#include "TimeDelta.h"
#include "URI.h"
#include "NetworkActivityReporter.h"
#include "Debug.h"
#include <stddef.h>
#include <cassert>

class ResponseErrorInitiator;
class DownloadProgress;

using namespace std;

class Client::ResponseHandler :
	public AbstractResponseHandler,
	public AbstractDataConsumer
{
	DECLARE_NON_COPYABLE(ResponseHandler)
public:
	ResponseHandler(
		ClientWeakPtr const& client_ptr, ConstRequestPtr const& request,
		RequestTag const& request_tag, long id, bool conn_close);
	
	virtual ~ResponseHandler();
private:
	virtual void processProvisionalResponse(std::auto_ptr<HttpResponseMetadata> metadata);
	
	virtual void processResponseMetadata(std::auto_ptr<HttpResponseMetadata> metadata);
	
	virtual void processBodyData(SplittableBuffer& data, bool eof);
	
	virtual void processError(IntrusivePtr<ErrorDescriptor> const& edesc);
	
	virtual void processNewData(SplittableBuffer& data, bool eof);
	
	bool tryDiscardOutput();
	
	ClientWeakPtr m_ptrClient;
	ConstRequestPtr const m_ptrRequest;
	RequestTag m_requestTag;
	long const m_id;
	bool m_isConnClose;
	bool m_isEOF;
	uintmax_t m_producedDataSize;
	HttpStreamWriter m_streamWriter;
};


class Client::ScopedSuspender
{
public:
	ScopedSuspender(Client& owner) : m_rOwner(owner) {
		m_rOwner.suspend();
	}
	
	~ScopedSuspender() {
		m_rOwner.resume();
	}
private:
	Client& m_rOwner;
};


Client::Client(ServiceContext& context, SocketPtr const& client_stream)
:	m_rContext(context),
	m_ptrClientStream(client_stream),
	m_isEndOfOutput(false),
	m_readTimer(
		sigc::mem_fun(*this, &Client::onReadTimeout),
		TimeDelta::fromSec(READ_TIMEOUT),
		&context.reactor()
	),
	m_writeTimer(
		sigc::mem_fun(*this, &Client::onWriteTimeout),
		TimeDelta::fromSec(WRITE_TIMEOUT),
		&context.reactor()
	),
	m_idleTimer(
		sigc::mem_fun(*this, &Client::onIdleTimeout),
		TimeDelta::fromSec(IDLE_TIMEOUT),
		&context.reactor()
	),
	m_eventMask(Reactor::READ),
	m_flags(0),
	m_isWaitingForWrite(0),
	m_responseIdCounter(0),
	m_lastResponseToProduceOutput(-1),
	m_ptrSelf(this)
{
	DBG_CLIENT_CONNECTION_BEGIN();
	m_ptrClientStream->enable(ACE_NONBLOCK);
	
	Reactor& reactor = m_rContext.reactor();
	reactor.beforeSleepSignal().connect(sigc::mem_fun(*this, &Client::updateState));
	try {
		m_handlerId = reactor.registerHandler(
			m_ptrClientStream->get_handle(),
			IntrusivePtr<EventHandlerBase>(this),
			m_eventMask
		);
	} catch (Reactor::Exception&) {
		DEBUGLOG("could not register client socket in the reactor");
		scheduleSessionTermination();
	}
}

Client::~Client()
{
	if (m_handlerId) {
		m_rContext.reactor().unregisterHandler(m_handlerId);
	}
	DBG_CLIENT_CONNECTION_END();
}

void
Client::processOutputData()
{
	//DEBUGLOG("Client::processOutputData()");
	while (!m_outputData.empty()) {
		if (writeOutputData() == 0) {
			break;
		}
	}
	
}

size_t
Client::writeOutputData()
{
	iovec iov[ACE_IOV_MAX];
	int iov_size = 0;
	for (SplittableBuffer::SpanIterator iter(m_outputData.spanBegin());
		iov_size < int(sizeof(iov)/sizeof(*iov)) && !iter.isAtRightBorder();
		++iov_size, ++iter) {
		iov[iov_size].iov_base = (char*)iter->begin();
		iov[iov_size].iov_len = iter->size();
	}
	ssize_t sent = m_ptrClientStream->sendv(
		iov, iov_size
		// the timeout is infinite, but the socket is in nonblocking mode
	);
	if (sent < 0 && errno == EWOULDBLOCK) {
		m_flags |= WRITE_BUFFER_FULL;
	} else if (sent > 0) {
		m_flags &= ~WRITE_BUFFER_FULL;
		m_writeTimer.postpone();
		NetworkActivityReporter::instance()->reportActivity();
#ifdef DEBUG
		SplittableBuffer traf;
		m_outputData.splitFrontAppendBack(sent, traf);
		DBG_TRAFFIC_TO_CLIENT(traf);
#else
		m_outputData.trimFront(sent);
#endif
		return sent;
	} else {
		DEBUGLOG("ERROR writing to client");
		scheduleSessionTermination();
	}
	return 0;
}

void
Client::processInputData(bool eof)
{
	//DEBUGLOG("Client::processInputData()");
	while (true) {
		if (m_inputData.empty() && !eof) {
			break;
		}
		if (isInactiveState()) {
			switchToRequestStart();
		}
		// at this point the parser is guaranteed to be in the RequestStart state
		SplittableBuffer body_data;
		bool body_eof = false;
		bool const empty_input = m_inputData.empty();
		bool const message_finishes = processNewData(m_inputData, eof, body_data, body_eof);
		if (message_finishes) {
			DBG_INCOMING_REQUEST_END();
		}
		if (!body_data.empty() || body_eof) {
			ScopedSuspender suspender(*this);
			m_ptrRequestHandler->appendBodyData(body_data, body_eof);
		}
		if (message_finishes) {
			m_ptrRequestHandler.reset(0);
			m_ptrRequestURI.reset(0);
		}
		if (isErrorState()) {
			DBG_INCOMING_REQUEST_ERROR();
			//DEBUGLOG("ERROR parsing request");
			IntrusivePtr<ErrorDescriptor> edesc(errorDescriptor());
			if (!edesc.get()) {
				edesc.reset(new ErrorDescriptor(
					ErrorCodes::ERROR_PARSING_RESPONSE,
					"error parsing request: "+errorMessage(),
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errParsingRequest),
						m_ptrRequestURI.get() ? m_ptrRequestURI->toString() : string(),
						errorMessage()
					)
				));
			}
			
			if (m_ptrRequestHandler.get()) {
				// the error happened after we've received the metadata
				m_ptrRequestHandler->cancelRequest(edesc);
				m_ptrRequestHandler.reset(0);
				m_ptrRequestURI.reset(0);
			} else {
				if (empty_input) {
					// an eof while in RequestStart state generates an error
					break;
				}
				// The error happened before we've received the metadata.
				// We make up a request and submit it with an error response attached.
				HttpRequestLine rline(BString("GET"), URI(BString()), HttpVersion::HTTP_1_0);
				RequestPtr request(new HttpRequestMetadata(rline));
				request->setBodyStatus(HttpRequestMetadata::BODY_FORBIDDEN);
				RequestTag request_tag;
				IntrusivePtr<ResponseHandler> response_handler(
					new ResponseHandler(
						m_ptrSelf, request, request_tag,
						++m_responseIdCounter, true /* conn_close */
					)
				);
				ScopedSuspender suspender(*this);
				IntrusivePtr<AbstractRequestHandler> request_handler(
					m_rContext.server().submitRequest(
						request, request_tag,
						response_handler, edesc
					)
				);
				SplittableBuffer body_data;
				request_handler->appendBodyData(body_data, true);
			}
			
			switchToInactive();
			m_flags |= READING_FORBIDDEN;
			m_inputData.clear();
			break;
		}
		if (!message_finishes) {
			// This means more input data is needed.
			// Example: the input ends in the middle of a header.
			break;
		}
	}
}

void
Client::processResponseData(SplittableBuffer& data, bool eof, long response_id)
{
	if (m_isEndOfOutput || isSessionTerminationScheduled()) {
		data.clear();
		return;
	}
	m_lastResponseToProduceOutput = response_id;
	m_isEndOfOutput = eof;
	m_outputData.appendDestructive(data);
	
	if (m_outputData.size() > MAX_WRITE_BUF_SIZE) {
		ScopedIncDec<int> flag_manager(m_isWaitingForWrite);
		while (!m_outputData.empty()) {
			if (m_rContext.reactor().handleEvents() != Reactor::SUCCESS) {
				break;
			}
		}
	}
}

bool
Client::tryDiscardResponseOutput(long response_id, uintmax_t discard_size)
{
	if (response_id != m_lastResponseToProduceOutput) {
		return false;
	}
	if (discard_size > m_outputData.size()) {
		return false;
	}
	m_outputData.trimBack(discard_size);
	return true;
}

bool
Client::isRequestInProgress() const
{
	return !(isInactiveState() || isInitialState() || isErrorState());
}

bool
Client::isIdle() const
{
	return (m_outputData.empty() && !isResponseInProgress() &&
		m_rContext.server().getNumPendingRequests() == 0 &&
		!isRequestInProgress());
}

void
Client::onReadTimeout()
{
	DEBUGLOG("client read timeout");
	scheduleSessionTermination();
}

void
Client::onWriteTimeout()
{
	DEBUGLOG("client write timeout");
	scheduleSessionTermination();
}

void
Client::onIdleTimeout()
{
	DEBUGLOG("client idle timeout");
	scheduleSessionTermination();
}

void
Client::terminateSession()
{
	m_readTimer.deactivate();
	m_writeTimer.deactivate();
	m_idleTimer.deactivate();
	m_eventMask = Reactor::NO_EVENTS;
	if (m_handlerId) {
		m_rContext.reactor().setEvents(m_handlerId, m_eventMask);
	}
	m_rContext.reactor().stop();
}

void
Client::updateState()
{
	/*
	upadateState() MUST NOT generate any output!!!
	Server::updateState() and Client::updateState() are called just before
	going to sleep to update their event masks. If one of them generates
	output for the other, that other one would need to update its event
	mask, but it may not have a chance to do so, as it may have already run
	in the current reactor dispatch.
	*/

	if (!m_outputData.empty() && !(m_flags & WRITE_BUFFER_FULL) &&
	    !isSessionTerminationScheduled()) {
		// necessary on the edge-triggered WFMOReactor
		processOutputData(); // can schedule session termination
	}
	if (m_rContext.server().getNumPendingRequests() == 0 && isReadingForbidden()) {
		m_isEndOfOutput = true;
	}
	if (m_outputData.empty() && m_isEndOfOutput) {
		scheduleSessionTermination();
	}
	if (isSessionTerminationScheduled()) {
		terminateSession();
		return;
	}
	
	Reactor::IOEvents mask = m_eventMask;
	
	if (m_flags & (READING_FORBIDDEN|READING_SUSPENDED) ||
	    m_rContext.server().getNumPendingRequests() >=
		m_rContext.server().getMaxPendingRequests()) {
		mask &= ~Reactor::READ;
	} else {
		mask |= Reactor::READ;
		
	}
	if (!m_outputData.empty()) {
		mask |= Reactor::WRITE;
	} else {
		mask &= ~Reactor::WRITE;
		if (m_isWaitingForWrite) {
			m_rContext.reactor().wakeup();
		}
		
		m_flags &= ~WRITE_BUFFER_FULL;
		// It may actually be full, but we don't want to miss
		// an edge-triggered (in case of WFMOReactor) write event.
	}
	
	if ((mask & Reactor::READ) && isRequestInProgress()) {
		m_readTimer.ensureActive();
	} else {
		m_readTimer.deactivate();
	}
	if (mask & Reactor::WRITE) {
		m_writeTimer.ensureActive();
	} else {
		m_writeTimer.deactivate();
	}
	if (isIdle()) {
		m_idleTimer.ensureActive();
	} else {
		m_idleTimer.deactivate();
	}
	
	if (mask != m_eventMask) {
		m_rContext.reactor().setEvents(m_handlerId, mask);
		m_eventMask = mask;
	}
}

void
Client::eventGotMetadata(auto_ptr<HttpRequestMetadata> metadata, bool is_persistent)
{
	m_ptrRequestURI.reset(new URI(metadata->requestLine().getURI()));
	
	DBG_REGISTER_CLIENT_REQUEST(*metadata);
	DBG_INCOMING_REQUEST_BEGIN(*metadata);
	
	HttpMessageShaper::shapeRequest(*metadata);
	
	ConstRequestPtr request(metadata.release());
	// metadata is null now !!!
	
	RequestTag request_tag;
	IntrusivePtr<ResponseHandler> response_handler(
		new ResponseHandler(
			m_ptrSelf, request, request_tag,
			++m_responseIdCounter, !is_persistent
		)
	);
	ScopedSuspender suspender(*this);
	m_ptrRequestHandler = m_rContext.server().submitRequest(
 		request, request_tag, response_handler
	);
}

void
Client::handleRead(ACE_HANDLE)
{
	if (m_flags & (READING_FORBIDDEN|READING_SUSPENDED)) {
		// this may happen because the event mask is updated after
		// all of the triggered events are processed
		return;
	}
	auto_ptr<DataChunk> chunk(DataChunk::create(READ_BUF_SIZE));
	ssize_t received = m_ptrClientStream->recv(
		chunk->getDataAddr(), chunk->getDataSize()
		// the timeout is infinite, but the socket is in nonblocking mode
	);
	if (received > 0) {
		NetworkActivityReporter::instance()->reportActivity();
		chunk = DataChunk::resize(chunk, received);
#ifdef DEBUG
		SplittableBuffer traf;
		traf.append(chunk);
		DBG_TRAFFIC_FROM_CLIENT(traf);
		m_inputData.appendDestructive(traf);
#else
		m_inputData.append(chunk);
#endif
	} else if (received < 0 && errno == EWOULDBLOCK) {
		// not a problem
		return;
	}
	m_readTimer.postpone();
	bool eof = (received <= 0);
	if (received < 0) {
		DEBUGLOG("ERROR reading from client");
	}
	processInputData(eof);
	if (received < 0 || (received == 0 && isIdle())) {
		scheduleSessionTermination();
	} else if (received == 0) {
		// an EOF from the client doesn't mean it's refusing to accept data
		m_flags |= READING_FORBIDDEN;
	}
}

void
Client::handleWrite(ACE_HANDLE)
{
	m_flags &= ~WRITE_BUFFER_FULL;
	processOutputData();
}


/*======================== Client::ResponseHandler ===========================*/

Client::ResponseHandler::ResponseHandler(
	ClientWeakPtr const& client_ptr, ConstRequestPtr const& request,
	RequestTag const& request_tag, long id, bool conn_close)
:	m_ptrClient(client_ptr),
	m_ptrRequest(request),
	m_requestTag(request_tag),
	m_id(id),
	m_isConnClose(conn_close),
	m_isEOF(false),
	m_producedDataSize(0),
	m_streamWriter(*this)
{
}

Client::ResponseHandler::~ResponseHandler()
{
}

void
Client::ResponseHandler::processProvisionalResponse(std::auto_ptr<HttpResponseMetadata> metadata)
{
	if (m_ptrRequest->requestLine().getHttpVersion() >= HttpVersion::HTTP_1_1) {
		m_streamWriter.startResponse(
			*metadata, m_ptrRequest->requestLine().getHttpVersion(), false
		);
		SplittableBuffer body_data;
		m_streamWriter.appendBodyData(body_data, true);
	}
}

void
Client::ResponseHandler::processResponseMetadata(std::auto_ptr<HttpResponseMetadata> metadata)
{
	if (Client* client = m_ptrClient.get()) {
		client->setResponseInProgress(true);
	}
	m_streamWriter.startResponse(
		*metadata, m_ptrRequest->requestLine().getHttpVersion(), m_isConnClose
	);
	m_isConnClose |= m_streamWriter.isConnCloseSet();

#ifdef DEBUG
	typedef AbstractDebugAgent ADA;
	ADA::HttpMessageType msg_type = ADA::UNMODIFIED_OUTGOING_RESPONSE;
	int const scode = metadata->statusLine().getCode();
	if (m_requestTag->flags().isSet(RequestTag::RESPONSE_CRAFTED)) {
		msg_type = ADA::CRAFTED_OUTGOING_RESPONSE;
	} else if (scode >= 400 && scode < 600) {
		msg_type = ADA::ERROR_OUTGOING_RESPONSE;
	} else if (m_requestTag->flags().isSet(RequestTag::RESPONSE_MODIFIED)) {
		msg_type = ADA::FILTERED_OUTGOING_RESPONSE;
	}
	DBG_HTTP_MESSAGE_BEGIN(*metadata, msg_type);
#endif
}

void
Client::ResponseHandler::processBodyData(SplittableBuffer& data, bool eof)
{
	if (m_isEOF) {
		data.clear();
		return;
	}
	m_isEOF = eof;
	if (m_isEOF) {
		DBG_OUTGOING_RESPONSE_END();
	}
	return m_streamWriter.appendBodyData(data, eof);
}

void
Client::ResponseHandler::processError(IntrusivePtr<ErrorDescriptor> const& edesc)
{
	DEBUGLOG(
		"Client::ResponseHandler::processError(): "
		<< edesc->getErrorMessage() << "\r\nURL: "
		<< m_ptrRequest->requestLine().getURI()
	);
	if (edesc->getErrorCode() == ErrorCodes::ERROR_DECOMPRESSING_RESPONSE) {
		// Some sites actually produce broken compressed content,
		// but browsers display it correctly, because the error
		// is at the end of file.
		if (m_producedDataSize != 0) {
			if (!m_isEOF) {
				m_isEOF = true;
				SplittableBuffer data;
				m_streamWriter.appendBodyData(data, true);
			}
		}
	}
	if (m_isEOF) {
		// ignore errors that come after EOF
		// that could be a REQUEST_CANCEL_TO_SAVE_TRAFFIC error
	} else if (m_producedDataSize == 0 || tryDiscardOutput()) {
		DBG_OUTGOING_RESPONSE_ERROR();
		m_streamWriter.reset();
		if (Client* client = m_ptrClient.get()) {
			if (client->isReadingForbidden()) {
				m_isConnClose = true;
			}
		}
		auto_ptr<CraftedResponse> response = edesc->responseGenerator()();
		response->metadata().setClientRequestId(m_ptrRequest->getClientRequestId());
#ifdef DEBUG
		response->metadata().headers().setHeader(
			BString("X-Request-URI"),
			m_ptrRequest->requestLine().getURI().toBString()
		);
#endif
		if (m_ptrRequest->requestLine().getMethod() == BString("HEAD")) {
			response->metadata().setBodyStatus(HttpResponseMetadata::BODY_FORBIDDEN);
			response->body().clear();
		}
		m_streamWriter.startResponse(
			response->metadata(),
			m_ptrRequest->requestLine().getHttpVersion(),
			m_isConnClose
		);
		DBG_ERROR_OUTGOING_RESPONSE_BEGIN(response->metadata());
		m_streamWriter.appendBodyData(response->body(), true);
		DBG_OUTGOING_RESPONSE_END();
	} else {
		//DEBUGLOG("some part of the response has already made its way to the client");
		DBG_OUTGOING_RESPONSE_ERROR();
		if (Client* client = m_ptrClient.get()) {
			client->scheduleSessionTermination();
		}
	}
}

void
Client::ResponseHandler::processNewData(SplittableBuffer& data, bool eof)
{
	m_producedDataSize += data.size();
	
	if (Client* client = m_ptrClient.get()) {
		bool stream_eof = eof && (m_isConnClose || client->isReadingForbidden());
		client->processResponseData(data, stream_eof, m_id);
		if (eof) {
			client->setResponseInProgress(false);
		}
	}
}

bool
Client::ResponseHandler::tryDiscardOutput()
{
	if (Client* client = m_ptrClient.get()) {
		if (client->tryDiscardResponseOutput(m_id, m_producedDataSize)) {
			m_producedDataSize = 0;
			return true;
		}
	}
	return false;
}
